#------------------------------------------------------------------------------
#	askmail.py - mail and mailbox related classes
#
#	(C) 2001-2002 by Marco Paganini (paganini@paganini.net)
#
#   This file is part of ASK - Active Spam Killer
#
#   ASK is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ASK is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ASK; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	$Id: askmail.py,v 1.20 2003/05/01 19:37:01 paganini Exp $
#------------------------------------------------------------------------------

import os
import re
import sys
import string
import rfc822
import time
import tempfile
import askversion
import asklog
import askconfig
import asklock
import whrandom

#------------------------------------------------------------------------------

class AskMail:
	"""
	This contains the mailbox delivery routines to be used by ASK.

	Attributes:
	
	- mailfrom: Sender
	- log:      ASKLOG object
	- config:   ASKCONFIG object

	Usage:

	config        = AskConfig.AskConfig(argv)
	log           = AskLog.AskLog("your_log_file")

	mail          = AskMail.AskMail(config, log)
	mail.mailfrom = "youremail@yourdomain.com"

	"""

	#------------------------------------------------------------------------------

	def __init__(self, config, log):
		"""
		Initializes the AskMail class
		"""

		self.config = config
		self.log    = log

		self.simple_headers = 0

	#------------------------------------------------------------------------------

	def	deliver_mail(self, mailbox, mailfile, custom_header="No further info", uniq = None):
		"""
		Deliver mail contained in 'mailfile' (no checks are done to make sure
		it's in mbox format) into mailbox. This function will append the contents
		of 'mailfile' to 'mailbox', doing fcntl style locking first. The function
		will try to deliver to a maildir style mailbox if the mailbox name ends
		with a "/". 'custom_header' will be appended to the list of headers.
		If "mailbox" is "-", stdout will be used. The 'uniq' parameter can be used
		to override the value of the 'uniq' field (the second field in a Maildir
		mailfile). If not set, the current uid will be used.
		"""

		self.log.write(1, "  deliver_mail(): Delivering mail from %s to mailbox %s" % (mailfile, mailbox))

		## If mailbox == "-", we use stdout. If it ends with a slash, it's
		## a Maildir mailbox otherwise we assume mbox format

		using_maildir = 0
		using_stdout  = 0

		lck = asklock.AskLock()

		## Stdout
		if (mailbox == "-"):
			using_stdout = 1
			mailbox = "stdout"		## Just for printing purposes
			self.log.write(5, "  deliver_mail(): delivering to stdout")
			fh_mbox = sys.stdout;

		## Maildir Format
		elif (mailbox[len(mailbox) - 1] == "/"):
			using_maildir = 1
			
			## Remove '/cur' '/new' and '/tmp' (jumps back to the root of the mailbox)
			mailbox = string.replace(mailbox, "/cur", "")
			mailbox = string.replace(mailbox, "/new", "")
			mailbox = string.replace(mailbox, "/tmp", "")

			if not uniq:
				uniq = os.getuid()

			## Try until the mailfile is not there. If it's there,
			## wait for a random amount of time and try again

			while 1:
				rnd              = whrandom.whrandom()	## Init the random generator
				filetime         = int(time.time())
				maildir_file     = os.path.join(mailbox, "tmp", "%ld.%s.%s" % (filetime, uniq, (os.uname())[1]))
				maildir_file_new = string.replace(maildir_file, "tmp/", "new/")

				if not (os.path.exists(maildir_file) or os.path.exists(maildir_file_new)):
					fh_mbox = lck.open(maildir_file, "w")
					break
				else:
					self.log.write(10, "  deliver_mail(): Maildir file exists. Sleeping and retrying...")
					time.sleep(int(rnd.random() * 3) + 1)

			self.log.write(5, "  deliver_mail(): Mailbox format: Maildir, tmp filename=%s" % maildir_file)

		## Mbox format
		else:
			self.log.write(5, "  deliver_mail(): Mbox format")

			lockf = ""
			if self.config.rc_lockfile != "":
				lockf = self.config.rc_lockfile + "." + os.path.basename(mailbox)

			fh_mbox = lck.open(mailbox, "a", lockf)
		
		## We suppose the file is unlocked when we get here...

		fh_mailfile = open(mailfile, "r")
		
		in_headers = 1;

		while 1:
			buf = fh_mailfile.readline();

			if (buf == ''):
				break

			## When the headers are exhausted (single crlf), we write our custom header
			if (len(string.strip(buf)) == 0 and in_headers):
				in_headers = 0
				fh_mbox.write("X-ASK-Info: %s (%s)\n" % (custom_header,time.strftime("%Y/%m/%d %H:%M:%S",time.localtime(time.time()))))

			fh_mbox.write(buf)
			lastbuf = buf

		fh_mailfile.close()
		
		## If lastbuf is not an empty line, we add one. mbox style messages
		## must always end in a newline

		if (len(string.strip(lastbuf)) != 0 and (not using_maildir)):
			fh_mbox.write("\n")

		## Unlock and close
		if not using_stdout:
			lck.close()

		## Rename mailbox from 'tmp' to 'new' directory
		if using_maildir:
			self.log.write(5, "  deliver_mail(): Renaming maildir file %s to %s\n" % (maildir_file, maildir_file_new))
			os.rename(maildir_file, maildir_file_new)

		self.log.write(1, "  deliver_mail(): Message delivered to %s" % mailbox)

		return 0

	#------------------------------------------------------------------------------

	def	send_mail(self, mailto, subject, body_filenames, attached_msgs = [],
		custom_headers = [], custom_footers = [], max_attach_lines=-1,
		html_mail = 0):
		"""
		This function will send mail as if coming from user 'self.mailfrom',
		to user 'mailto' using the subject 'subject'. The body of the
		mail will be retrieved from the given 'body_filenames' and
		'attached_msgs'. Only From:,Subject,To,Date, Cc and Bcc will be
		preserved if self.config.rc_basic_headers set. 'max_attach_lines' tells
		how many lines to send back in the attachment. 'custom_headers' allows
		you to specify an array of custom headers to be added to the message
		header. If html_mail is set to 1, the proper headers will be generated
		and the 'copyright' footer will be presented in html format.
		"""

		self.log.write(10, "  send_mail(): mailfrom=%s, mailto=%s, subject=%s" %
			(self.mailfrom, mailto, subject))

		temp = "%s.%d" % (tempfile.mktemp(), os.getpid())
		self.log.write(10, "  send_mail(): tempfile=" + temp)

		fh = open(temp, "w")

		## Create 'From:' line
		if (self.fullname != ''):
			fromline = 'From: "%s" <%s>' % (self.fullname, self.mailfrom)
		else:
			fromline = 'From: %s' % self.mailfrom
			
		fh.write(fromline    + "\n")
		fh.write("To: "      + mailto   + "\n")
		fh.write("Subject: " + subject  + "\n")
		fh.write("X-AskVersion: " + askversion.AskVersion.version + " (http://www.paganini.net/ask)\n")

		## Process HTML Mail

		if html_mail:
			fh.write("Content-Type: text/html; charset=\"iso-8859-1\"\n")
			fh.write("Mime-Version: 1.0\n")

		## Write the custom headers
		for var in custom_headers:
			fh.write(var + "\n")

		fh.write("\n")			## Header / Message separator

		## Write non RFC-822 bodies

		for var in body_filenames:
			self.log.write(10, "  send_mail(): body_filename=%s" % var)

			fbody = open(var, "r")

			while 1:
				buf = fbody.readline();
				if (buf == ''):
					break

				## Quote "From:" Lines
				if (buf[0:5] == 'From '):
					buf = ">" + buf

				fh.write(buf)

			fh.write("\n")		## Separator
			fbody.close()

		## Shameless copyrights

		if html_mail:
			fh.write("<p><font size=\"-1\">\n")
			fh.write("This email account is protected by:<br>\n")
			fh.write("<a href=\"http://www.paganini.net/ask\">Active Spam Killer</a> (ASK) V%s - (C) 2001-2002 by Marco Paganini<br>\n" % askversion.AskVersion.version)
			fh.write("For more information visit <a href=\"http://www.paganini.net/ask\">http://www.paganini.net/ask<a>\n")
			fh.write("</font><p>")
		else:
			fh.write("This email account is protected by:\n")
			fh.write("Active Spam Killer (ASK) V%s - (C) 2001-2002 by Marco Paganini\n" % askversion.AskVersion.version)
			fh.write("For more information visit http://www.paganini.net/ask\n\n")

		## Only write separator *IF* there's something to be added at the end

		if len(attached_msgs):
			if html_mail:
				fh.write("<b>Original Message Follows:</b>\n<hr>\n")
			else:
				fh.write("--- Original Message Follows ---\n\n")

		## Write RFC-822 bodies, mangling headers if necessary

		for var in attached_msgs:
			self.log.write(10, "  send_mail(): attached_msgs=%s" % var)

			fbody = open(var, "r")

			in_headers = 1

			## We only send max_attach_lines (if not -1)
			max_lines = max_attach_lines

			if html_mail:
				print "<pre>\n"

			while (max_lines >= 0 or max_attach_lines == -1):

				buf = fbody.readline();
				if (buf == ''):
					break

				## End of headers?
				if len(string.strip(buf)) == 0:
					in_headers = 0

				## Mangle, if necessary
				if (self.config.rc_basic_headers != [] and in_headers):
					hdr = string.split(buf)[0]				## First word
					if hdr not in self.config.rc_basic_headers:
						continue;
				
				## Quote "From:" Lines in any_case
				if (buf[0:5] == 'From '):
					buf = ">" + buf

				fh.write(buf)

				## Headers do not count in the maximum number of lines in attachment
				
				if not in_headers:
					max_lines = max_lines - 1

			## Indicate the original was truncated if needed

			if (max_attach_lines != -1 and max_lines <= 0):
				fh.write("\n(Original message truncated)\n")

			if html_mail:
				print "</pre><p>\n"
			else:
				fh.write("\n")

		## Write custom footers
		for var in custom_footers:
			fh.write(var + "\n")

		fh.write("\n")		## Separator

		fh.close()

		## Call sendmail to send it

		command = string.replace(self.config.rc_mta_command, "MAILFILE", temp)

		ret = os.system(command)
		self.log.write(10, "send_mail(): os.system(%s) returned %d" % (command, ret))
		os.unlink(temp)

	#------------------------------------------------------------------------------

	def	send_mailfile(self, mailto, mailfilename, custom_headers = []):
		"""
		This function will invoke sendmail on the specified 'mailfilename'.
		It will also add the 'custom_headers' to the list of headers in
		the message.
		"""

		self.log.write(10, "  send_mailfile(): mailto=%s, mailfilename=%s" %
			(mailto, mailfilename))

		temp = "%s.%d" % (tempfile.mktemp(), os.getpid())
		self.log.write(10, "  send_mail(): tempfile=" + temp)

		fread  = open(mailfilename, "r")
		fwrite = open(temp, "w")

		## Copy 'mailfilename' RFC822 headers to 'temp', eliminating
		## any old X-ASK-Auth headers it finds and adding a new one.

		while 1:
			buf = fread.readline()
			if (buf == ''):
				break

			## End of headers? Write custom headers and proceed to body
			if len(string.strip(buf)) == 0:
				for var in custom_headers:
					fwrite.write(var + "\n")
				fwrite.write("\n")			## Header / Message separator
				break

			## Copy everything else (except old X-ASK-Auth's)
			if not re.search("^X-ASK-Auth:", buf, re.IGNORECASE):
				fwrite.write(buf)

		## Write body

		while 1:
			buf = fread.readline()
			if (buf == ''):
				break

			fwrite.write(buf)

		fwrite.write("\n")		## Just to be sure there will be a separator...
		fread.close()
		fwrite.close()

		## QUIRK: To avoid adding another line just for a different syntax
		## of sendmail, we resort to using the first argument in rc_mta_command
		## (hopefully the executable name) directly.

		command = string.split(self.config.rc_mta_command)[0] + " " + mailto + " <" + temp

		ret = os.system(command)
		self.log.write(10, "send_mailfile(): os.system(%s) returned %d" % (command, ret))
		os.unlink(temp)

## EOF ##
