#------------------------------------------------------------------------------
#	askmessage.py
#
#	(C) 2001-2002 by Marco Paganini (paganini@paganini.net)
#
#   This file is part of ASK - Active Spam Killer
#
#   ASK is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ASK is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ASK; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	$Id: askmessage.py,v 1.80 2003/05/19 14:17:22 paganini Exp $
#------------------------------------------------------------------------------

import os
import os.path
import sys
import string
import rfc822
import md5
import re
import time
import tempfile
import mimify
import asklog
import askconfig
import askmail
import asklock

#------------------------------------------------------------------------------

class AskMessage:
	"""
	This is the main ASK class. 
	
	Attributes:

	- tmpfile:       Temporary file containing the mail message
	- fh:            File handle to the temporary file (do a seek(0) before using)
	- binary_digest: MD5 digest in binary format
	- ascii_digest:  MD5 digest in ascii format
	- conf_md5:      MD5 digest found in the subject

	- msg:	   rfc822 object
	- md5sum:  md5sum object
	- log:     LOG object
	- config:  CONFIG object
	- mail:    MAIL object
	"""

	#----------------------------------------------------------------------------------
	def __init__(self, config, log):
		"""
		Initializes the class instance. We need at least a valid CONFIG object 
		and a valid LOG object.
		"""

		self.fh = ''
		self.tmpfile = ''
		self.binary_digest = '';
		self.ascii_digest  = '';
		self.list_match    = '';
		self.set_conf_md5('');

		## Initialize the LOG, CONFIG and MAIL objects

		self.config        = config
		self.log           = log
		self.mail          = askmail.AskMail(self.config, self.log)

		self.mail.fullname = self.config.rc_myfullname;

	#----------------------------------------------------------------------------------
	def __del__(self):
		if self.fh:
			self.fh.close();

		#self.log.write(10, "  __del__(): Removing %s" % self.tmpfile)
		os.unlink(self.tmpfile)

	#----------------------------------------------------------------------------------
	def set_conf_md5(self, md5):
		"""
		Allows external callers to change the notion of the 'conf_md5',
		or the md5sum retrieved from the subject line. This is useful
		for instance, when you wish to process different (queued) files
		using the methods in this class.
		"""

		self.conf_md5 = md5
		
	#----------------------------------------------------------------------------------
	def	read(self, input_fh):
		"""
		This function will read all data from the passed 'input_fh' into a
		temporary file. This file will then be used for all sorts of operations
		needed by this class. 
		"""

		## tmpfile must be on the same filesystem as ASK's private dir!
		## We'll use 'rename' later to change it to the definitive location
		self.tmpfile = "%s/%s.%d.msg" % (self.config.rc_tmpdir, os.path.basename(tempfile.mktemp()), os.getpid())
		
		## Create the temp filename and reopen as read/write
		self.fh = open(self.tmpfile, "w")
		self.fh.close()
		self.fh = open(self.tmpfile, "r+b")

		## Create a new MD5 object
		self.md5sum = md5.new()

		## Copy all the data from the passed file object to the new file
		
		while 1:
			buf = input_fh.readline()
			if (buf == ''):
				break
	
			self.md5sum.update(buf)
			self.fh.write(buf)
	
		## Add the md5key to the hash and produce the ASCII digest

		if (self.config.rc_md5_key != ''):
			self.md5sum.update(self.config.rc_md5_key)

		self.ascii_digest = ''
		binary_digest = self.md5sum.digest()

		for ch in range(0,len(binary_digest)):
			self.ascii_digest = self.ascii_digest + "%02.2x" % ord(binary_digest[ch])
		
		## Initialize the rfc822 object for our workfile

		self.fh.seek(0)
		self.msg = rfc822.Message(self.fh, 1)

	#----------------------------------------------------------------------------------
	def	get_real_sender(self):
		"""
		This is just like "get_sender", but the "X-Primary-Address" is not taken into
		account. Useful for cases where you don't want to use this address (when sending
		an email back, for instance).
		"""
		
		return(self.get_sender(ignore_header = "X-Primary-Address"))

	#----------------------------------------------------------------------------------
	def get_sender(self, ignore_header=""):
		"""
		Returns a tuple in the format (sender_name, sender_email). Some processing is
		done to treat X-Primary-Address/Resent-From/Reply-To/From. Headers in "ignore_header"
		are not checked.
		"""

		headers = ["X-Primary-Address", "Resent-From", "Reply-To", "From"];

		## Remove unwanted headers
		def fn_remove(x, y = ignore_header): return(x != y)
		headers = filter(fn_remove, headers)

		sender_name = ''
		sender_mail = ''

		for hdr in headers:
			(sender_name, sender_mail) = self.msg.getaddr(hdr)

			if sender_name == None: sender_name = ''
			if sender_mail == None: sender_mail = ''

			## Decode sender name and return if found
			if sender_mail:
				if sender_name:
					sender_name = mimify.mime_decode_header(sender_name)
				break

		return(string.rstrip(sender_name), string.rstrip(sender_mail))

	#----------------------------------------------------------------------------------
	def get_recipients(self):
		"""
		Returns a list of tuples in the format (sender_name, sender_email). Note that
		the list includes the contents of the TO, CC and BCC fields.
		"""
		
		to_list  = self.msg.getaddrlist("To")		## To:
		cc_list  = self.msg.getaddrlist("Cc")		## Cc:
		bcc_list = self.msg.getaddrlist("Bcc")		## Bcc:

		recipients = []

		## Master list of recipients

		recipients.extend(to_list)
		recipients.extend(cc_list)
		recipients.extend(bcc_list)

		return(recipients)

	#----------------------------------------------------------------------------------
	def get_subject(self):
		"""
		Returns the subject of the message, or '' if none is defined.
		"""

		return(mimify.mime_decode_header(self.msg.getheader("Subject", "")))

	#----------------------------------------------------------------------------------
	def get_date(self):
		"""
		Returns a tuple containing the broken down date (from the "Date:" field)
		The tuple contains (year, month, day, hour, minute, second, weekday, julianday,
		dst_flag) and can be fed directly to time.mktime or time.strftime.
		"""

		## Lots of messages have broken dates
		try:
			tuple = rfc822.parsedate(self.msg.getheader("Date", ""))
		except:
			tuple = ''

		return(tuple)

	#------------------------------------------------------------------------------
	def get_message_id(self):
		"""
		Returns the "Message-Id" field or '' if none is defined.
		"""

		return(self.msg.getheader("Message-Id", ""))

	#------------------------------------------------------------------------------
	def __inlist(self, filenames, sender=None, recipients=None, subj=None):
		"""
		Checks if sender email, recipients or subject match one of the regexps
		in the given filename array/tuple. If sender/recipient/subject are not
		specified, the defaults for the current instance will be used.
		"""

		## Fill in defaults
		if sender == None:
			sender = self.get_sender()[1]

		if recipients == None:
			recipients = self.get_recipients()

		if subj == None:
			subj = self.get_subject()

		## If the sender is one of our emails, we immediately return false.
		## This allows the use of "from @ourdomain.com" without having to
		## worry about matching from ouremail@ourdomain.com

		if self.is_from_ourselves(sender):
			self.log.write(10, "  __inlist(): We are the sender (%s). No match performed." % sender)
			self.list_match = ""
			return 0

		## Test each filename in turn

		for fname in filenames:

			## If the file does not exist, ignore
			self.log.write(5, "  __inlist(): filename=%s, sender=%s, dest=%s, subject=%s"% (fname, sender, recipients, subj))

			if (os.access(fname, os.R_OK) == 0):
				self.log.write(5, "  __inlist(): %s does not exist (or is unreadable). Ignoring..." % fname)
				continue
			
			fh = open(fname, "r")

			while 1:

				buf = fh.readline()
				if (buf == ''):
					break
			
				buf = string.lower(string.rstrip(buf))

				## Ignore comments and blank lines
				if (buf != '' and buf[0] != '#'):
					## self.log.write(20,"  __inlist(): regexp=[%s]" % buf)

					## "from re"    matches sender
					## "to re"      matches destination (to, bcc, cc)
					## "subject re" matches subject
					## "header re"  matches a full regexp in the headers
					## "re"         matches sender (previous behavior)

					strlist = []

					## from
					res = re.match("from\s*(.*)", buf, re.IGNORECASE)
					if res:
						## No blank 'from'
						if not res.group(1):
							continue

						regex   = self.__regex_adjust(res.group(1))
						strlist.append(sender)
					else:
						## to
						res = re.match("to\s*(.*)", buf, re.IGNORECASE)
						if res:
							regex = self.__regex_adjust(res.group(1))
							## Append all recipients
							for (recipient_name, recipient_email) in recipients:
								strlist.append(recipient_email)
						else:
							## subject
							res = re.match("subject\s*(.*)", buf, re.IGNORECASE)

							if res:
								regex = res.group(1)
								strlist.append(subj)
							else:
								## header
								res = re.match("header\s*(.*)", buf, re.IGNORECASE)
								if res:
									regex = res.group(1)
									strlist.extend(self.msg.headers)
								else:
									regex = self.__regex_adjust(buf)
									strlist.append(sender)

					## Match...

					for str in strlist:
						try:
							if (regex and re.search(regex, str, re.IGNORECASE)):
								self.log.write(5, "  __inlist(): found with re=%s" % buf)
								self.list_match = buf
								fh.close()
								return 1
						except:
							self.log.write(1,"  __inlist(): WARNING: Invalid regular expression: \"%s\". Ignored." % regex)
				else:
					self.log.write(20, "  __inlist(): ignoring line [%s]" % buf)

			fh.close()
		
		self.list_match = ""
		return 0

	#------------------------------------------------------------------------------
	def __regex_adjust(self, str):
		"""
		Returns the "adjusted" regular expression to be used in matching
		email items. If the regular expression itself matches xxx@xxx, we
		assume a full address match is desired (we try to match using ^re$ or
		else r@boo.org would also match spammer@boo.org). Otherwise, just use
		a regular regexp to match anywhere.
		"""

		if re.search(".+@.+", str):
			return("^" + str + "$")
		else:
			return(str)

	#------------------------------------------------------------------------------
	def is_from_ourselves(self, email = None):
		"""
		Returns true if the sender is one of our emails. Use 'email' instead of
		the current sender if specified.
		"""

		if not email:
			email = string.lower(self.get_sender()[1])

		if (email in map(string.lower, self.config.rc_mymails)):
			return 1
		else:
			return 0
	
	#------------------------------------------------------------------------------
	def is_in_blacklist(self):
		"""
		Returns true if this message is in the blacklist. False otherwise
		"""
		if self.__inlist(self.config.rc_blacklist):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def is_in_whitelist(self):
		"""
		Returns true if this message is in the whitelist. False otherwise
		"""
		if self.__inlist(self.config.rc_whitelist):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def is_in_ignorelist(self):
		"""
		Returns true if this message is in the ignorelist. False otherwise
		"""
		if self.__inlist(self.config.rc_ignorelist):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def has_our_key(self):
		"""
		Returns true if this message contains our mailkey. False otherwise
		"""
		if self.__contains_string(self.config.rc_mailkey):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def is_from_mailerdaemon(self):
		"""
		Returns true if this message comes from MAILER-DAEMON. False otherwise
		"""

		if re.match("mailer-daemon|postmaster@|<>", self.get_sender()[1], re.IGNORECASE):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def sent_by_ask(self):
		"""
		Looks for the X-AskVersion rfc822 header. If one is found, we assume ASK
		sent the message and we return true. Otherwise, we return false.
		"""

		if self.__contains_string("^X-AskVersion:"):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def	deliver_mail(self, custom_header="Delivering Mail"):
		"""
		Delivers the current mail to the mailbox contained in self.config.rc_mymailbox
		"""

		self.log.write(1, "  deliver_mail(): Delivering mail")
		self.mail.deliver_mail(self.config.rc_mymailbox, self.tmpfile, custom_header)

		return 0

	#------------------------------------------------------------------------------
	def	junk_mail(self, custom_header="Junk"):
		"""
		Delivers the current mail to the junkmailbox (or discards if rc_junkmailbox
		is undefined)
		"""

		self.log.write(1, "  junk_mail(): Queueing Junk Message")
		self.queue_mail(custom_header = custom_header)
		return 0

	#------------------------------------------------------------------------------
	def	bulk_mail(self, custom_header="Bulk"):
		"""
		Delivers the current mail to the bulkmailbox (or discards if rc_bulkmailbox
		is undefined)
		"""

		self.log.write(1, "  bulk_mail(): Queueing Bulk Message")
		self.queue_mail(custom_header = custom_header)
		return 0

	#------------------------------------------------------------------------------
	def queue_mail(self, custom_header="Message Queued"):
		"""
		Queues the current message. The queue directory and calculated MD5
		are used to generate the filename. Note that the MD5 reflect the original
		contents of the message (not the queued one, as headers are added as needed).

		If the message queue directory (rc_askmsg) ends in "/", we assume the
		queue will be in mqueue format.
		"""

		self.log.write(5,  "  queue_mail(): custom_header = %s" % custom_header)
		self.log.write(5,  "  queue_mail(): The MD5 checksum for %s is %s" % (self.tmpfile, self.ascii_digest))

		## If the msgdir ends in "/" we assume a Maildir style queue. If
		## not, we have a regular queue.

		if self.config.queue_is_maildir:
			self.log.write(5, "  queue_mail(): Maildir format. Queue dir = %s" % self.config.rc_msgdir)
			self.mail.deliver_mail(self.config.rc_msgdir, self.tmpfile, custom_header, uniq = self.ascii_digest)
		else:
			queueFileName = self.queue_file()
			self.log.write(5, "  queue_mail(): Mailbox format. Queue file = %s" % queueFileName)
			self.mail.deliver_mail(queueFileName, self.tmpfile, custom_header)
		
		return 0

	#------------------------------------------------------------------------------
	def send_confirmation(self):
		"""
		Sends a confirmation message back to the sender.
		"""

		queueFileName = self.queue_file()

		self.log.write(1, "  send_confirmation(): Sending confirmation to %s" % self.get_real_sender()[1])

		self.mail.mailfrom = self.__match_recipient()

		## Add Precedence: bulk and (if appropriate) "In-Reply-To"

		header_list = [ "X-ASK-Auth: %s" % self.generate_auth(), "Precedence: bulk", "Content-Type: text/plain; charset=\"iso-8859-1\"", "Content-Transfer-Encoding: 8bit" ]

		msgid = self.get_message_id()

		if msgid:
			header_list.append("In-Reply-To: %s" % msgid)
		
		if (self.__check_confirm_list(self.get_real_sender()[1])):
			self.mail.send_mail(self.get_real_sender()[1],
							"Please confirm (conf#%s)" % self.ascii_digest,
							self.config.rc_confirm_filenames,
							[queueFileName],
							custom_headers = header_list,
							max_attach_lines = self.config.rc_max_attach_lines)

			self.log.write(1, "  send_confirmation(): Confirmation sent to %s..." % self.get_real_sender()[1])
		else:
			self.log.write(1, "  send_confirmation(): too many confirmations already sent to %s..." % self.get_real_sender()[1])

			tempFile = "%s.%d" % (tempfile.mktemp(), os.getpid())
			tempFileHandle = open(tempFile, "w")
			tempFileHandle.write("""
ASK Warning:

You've received more than %d messages coming from the same sender in a short
period of time. To avoid mail loops, ASK will stop sending mail confirmations
to this address for a while. This situation is normally caused by poorly
configured auto-responders. You may want to check the original message
(below) that caused this problem.
			""" % (self.config.rc_max_confirmations))

			tempFileHandle.close()

			self.mail.send_mail(self.config.rc_mymails[0],
								"ASK Warning: Runaway Confirmations!",
								[tempFile],
								[queueFileName],
								custom_headers = header_list,
								max_attach_lines = self.config.rc_max_attach_lines)

			os.unlink(tempFile)

	#------------------------------------------------------------------------------

	def	__check_confirm_list(self, address):
		"""
		Adds address to the list of confirmation targets.
		Returns true if this confirmation should be sent.
		Returns false otherwise.
		"""
		
		name = address + "\n"

		queue_file_name = "%s/.ask-loop-control.dat"  % self.config.rc_askdir

		## Confirmation list file will be created if it does not exist
		if not os.path.exists(queue_file_name):
			queue_file_handle = open(queue_file_name, "w")
			queue_file_handle.close()

		## Lock
		lck   = asklock.AskLock()
		lockf = ""

		if self.config.rc_lockfile != "":
			lockf = self.config.rc_lockfile + "." + os.path.basename(queue_file_name)
		else:
			lockf = queue_file_name

		queue_file_handle = lck.open(queue_file_name, "r+", lockf)

		## Read all stuff from .ask-loop-control.dat

		if os.path.exists(queue_file_name):
			queue_file_handle.seek(0)
			confirm_list = queue_file_handle.readlines()
		else:
			confirm_list = []

		confirm_list.append(name)

		## Trim list
		if len(confirm_list) > self.config.rc_max_confirmation_list:
			del confirm_list[0:self.config.rc_max_confirmation_list - self.config.rc_min_confirmation_list]

		## Rewrite trimmed list and release lock

		queue_file_handle.seek(0)
		queue_file_handle.truncate(0)
		queue_file_handle.writelines(confirm_list)

		lck.close()

		## More than max_confirmations?
		if (confirm_list.count(name) > self.config.rc_max_confirmations):
			return 0

		return 1
	#------------------------------------------------------------------------------

	def	queue_file(self, md5str=""):
		"""
		Returns the full path to the queue file of the current message
		(self.ascii_digest). The 'md5' parameter can be used to override
		the MD5 checksum. If the queue is in maildir format, the queue
		directory will be opened and the first file containing the MD5
		will be returned.
		"""
		
		if not md5str:
			md5str = self.ascii_digest

		if self.config.queue_is_maildir:

			## Original Path if Maildir
			maildir = string.replace(self.config.rc_msgdir, "/cur", "")
			maildir = string.replace(maildir, "/new", "")
			maildir = string.replace(maildir, "/tmp", "")

			## /cur
			file_list = filter(lambda x,md5str_in = md5str: (string.find(x, md5str_in) != -1), os.listdir(maildir + "cur"))

			if len(file_list) != 0:
				return(os.path.join(maildir + "cur", file_list[0]))

			## /new
			file_list = filter(lambda x,md5str_in = md5str: (string.find(x, md5str_in) != -1), os.listdir(maildir + "new"))

			if len(file_list) != 0:
				return(os.path.join(maildir + "new", file_list[0]))

			## Nothing found, return blank
			return ""

		else:
			return "%s/ask.msg.%s" % (self.config.rc_msgdir, md5str)

	#------------------------------------------------------------------------------

	def	dequeue_mail(self, custom_header="Message dequeued", use_smtp = 0, mailbox = ''):
		"""
		Dequeues (delivers) mail in the queue directory to the current user.
		If 'use_smtp' is set, sendmail will be used to do the delivery. Otherwise,
		the queued message will be directly appended to the mailbox. The 'mailbox'
		parameter can be specified to force delivered to something different than
		the default config.rc_mymailbox.
		"""

		if not mailbox:
			mailbox = self.config.rc_mymailbox

		queueFileName = self.queue_file(self.conf_md5)

		self.log.write(1, "  dequeue_mail(): Delivering mail from %s to mailbox %s" % (queueFileName, mailbox))

		if use_smtp:
			self.mail.mailfrom = self.config.rc_mymails[0]

			self.mail.send_mailfile(self.config.rc_mymails[0],
									queueFileName,
									custom_headers = [
										"X-ASK-Auth: " + self.generate_auth(),
										"X-ASK-Info: " + custom_header ])
		else:
			self.mail.deliver_mail(mailbox,
								   queueFileName,
								   custom_header)

		os.unlink(queueFileName)

		return 0

	#------------------------------------------------------------------------------

	def	delete_mail(self, custom_header="No further info"):
		"""
		Deletes queued file in the queue directory.
		"""

		queueFileName = self.queue_file(self.conf_md5)
		os.unlink(queueFileName)

		self.log.write(1, "  delete_mail(): Queued file %s deleted" % queueFileName)

		return 0

	#------------------------------------------------------------------------------
	def	is_queued(self):
		"""
		Checks if a queued message exists matching the current MD5 signature.
		"""

		queueFileName = self.queue_file()

		if (os.access(queueFileName, os.F_OK) == 1):
			self.log.write(1, "  is_queued(): File %s found. Message is queued" % queueFileName)
			return 1
		else:
			self.log.write(1, "  is_queued(): File %s not found. Message is not queued" % queueFileName)
			return 0

	#------------------------------------------------------------------------------
	def	confirmation_msg_queued(self):
		"""
		Returns true if the current message is a confirmation message AND
		a queued message exists. False otherwise.
		"""

		queueFileName = self.queue_file(self.conf_md5)

		if (os.access(queueFileName, os.F_OK) == 1):
			self.log.write(1, "  confirmation_msg_queued(): File %s found. Message is queued" % queueFileName)
			return 1
		else:
			self.log.write(1, "  confirmation_msg_queued(): File %s not found. Message is not queued" % queueFileName)
			return 0

	#------------------------------------------------------------------------------
	def send_nastygram(self):
		"""
		Sends a nastygram back.
		"""

		self.log.write(1, "  send_nastygram(): Sending nastygram back")

		self.mail.mailfrom = self.__match_recipient()

		self.mail.send_mail(self.get_real_sender()[1],
			"Please stop sending me emails", 
			self.config.rc_blacklist_filenames,
			[self.tmpfile],
			custom_headers = [ "X-ASK-Auth: %s" % self.generate_auth(), "Precedence: bulk" ])

		return 0

	#------------------------------------------------------------------------------
	def is_confirmation_return(self):
		"""
		Checks whether the message subject is a confirmation. If so, self.conf_md5
		will be set to the MD5 hash found in the subject true will be returned.
		"""

		subject = self.get_subject()

		self.log.write(10, "  is_confirmation_return(): Subject=" + subject)

		res = re.search("\(conf[#:]([a-f0-9]{32})\)", subject, re.IGNORECASE)

		if (res == None):
			self.log.write(1, "  is_confirmation_return(): Didn't find conf#MD5 tag on subject")
			self.conf_md5 = ''
			return 0
		else:
			self.log.write(1, "  is_confirmation_return(): Found conf$md5 tag, MD5=%s" % res.group(1))
			self.conf_md5 = res.group(1)
			return 1

	#------------------------------------------------------------------------------
	def add_queued_msg_to_whitelist(self):
		"""
		Adds the sender in the message pointed to by 'conf_md5' to the whitelist.
		"""

		queueFileName   = self.queue_file(self.conf_md5)

		## Create a new AskMessage instance with the queued file
		aMessage        = AskMessage(self.config, self.log)
		queueFilehandle = open(queueFileName, "r")

		aMessage.read(queueFilehandle)
		queueFilehandle.close()

		self.log.write(1, "  add_queued_msg_to_whitelist(): Adding message %s to whitelist" % queueFileName)
		aMessage.add_to_whitelist()

	#------------------------------------------------------------------------------
	def add_to_whitelist(self):
		"""
		Adds the current sender to the whitelist. 
		"""

		if not self.is_in_whitelist():
			self.__add_to_list(self.config.rc_whitelist, self.get_sender()[1])
		else:
			self.log.write(5, "  add_to_whitelist(): Sender already in whitelist. Skipping.")

		return 0

	#------------------------------------------------------------------------------
	def add_to_ignorelist(self):
		"""
		Adds the current sender to the ignorelist. 
		"""

		if not self.is_in_ignorelist():
			self.__add_to_list(self.config.rc_ignorelist, self.get_real_sender()[1])
		else:
			self.log.write(5, "  add_to_ignorelist(): Sender already in ignorelist. Skipping.")

		return 0

	#------------------------------------------------------------------------------
	def add_to_blacklist(self):
		"""
		Adds the current sender to the blacklist
		"""

		if not self.is_in_blacklist():
			self.__add_to_list(self.config.rc_blacklist, self.get_real_sender()[1])
		else:
			self.log.write(5, "  add_to_blacklist(): Sender already in blacklist. Skipping.")

		return 0

	#------------------------------------------------------------------------------
	def invalid_sender(self):
		"""
		Performs some basic checking on the sender's email and return true if
		it seems to be invalid.
		"""

		sender_email = self.get_sender()[1]

		if (sender_email == '' or string.find(sender_email, '@') == -1 or string.find(sender_email,'|') != -1):
			self.log.write(1, "  invalid_sender(): Malformed 'From:' line (%s). " % sender_email)
			return 1
		else:
			return 0

	#-----------------------------------------------------------------------------
	def __add_to_list(self, filename, email=None):
		"""
		Adds the specified 'email' to the first filename in the array 'filename'.
		Defaults to using the current sender if none is specified.
		"""

		## Defaults
		if email == None:
			email = self.get_sender()[1]

		## Make sure it's not one of our own emails...
		if self.is_from_ourselves(email):
			self.log.write(1, "  __add_to_list(): \"%s\" is listed as one of our emails. It will not be added to the list (%s)" % (email, filename[0]))
			return -1

		self.log.write(10, "  __add_to_list(): filename=%s, email=%s" % (filename[0], email))

		lck = asklock.AskLock()

		lockf = ""
		if self.config.rc_lockfile != "":
			lockf = self.config.rc_lockfile + "." + os.path.basename(filename[0])

		fh = lck.open(filename[0], "a", lockf)
		
		## We suppose the file is unlocked when we get here...
		fh.write("from " + self.__escape_regex(email) + "\n")
		lck.close()

		return 0

	#------------------------------------------------------------------------------
	def __escape_regex(self, str):
		"""
		Escapes dots and other meaningful characters in the regular expression.
		Returns the "escaped" string. This routine is pretty dumb meaning that
		it does not know how to handle an already escaped string. Use only in
		"raw", unescaped strings.
		"""
		
		evil_chars = "\\.^$*+|{}[]";

		for var in evil_chars:
			str = string.replace(str, var, "\\"+var)

		return(str)

	#------------------------------------------------------------------------------
	def is_mailing_list_message(self):
		"""
		We try to identify (using some common headers) whether this message comes
		from a mailing list or not. If it does, we return 1. Otherwise, return 0
		"""

		if (self.msg.getheader("Mailing-List",'') != '' or
			self.msg.getheader("List-Id",'')      != '' or
			self.msg.getheader("List-Help",'')    != '' or
			self.msg.getheader("List-Post",'')    != '' or
			self.msg.getheader("Return-Path",'')  == '<>' or
			re.match("list|bulk", self.msg.getheader("Precedence",''), re.IGNORECASE) or
			re.match(".*(majordomo|listserv|listproc|netserv|owner|bounce|mmgr|autoanswer|request|noreply|nobody).*@", self.msg.getheader("From",''), re.IGNORECASE)):
			return 1
		else:
			return 0

	#------------------------------------------------------------------------------
	def generate_auth(self):
		"""
		Generates an authentication string containing the current time and the
		MD5 sum of the current time + our rc5_key. This is normally sent out
		in every email generated by ASK in the X-ASK-Auth: email header.
		"""

		md5sum = md5.new()

		## number of seconds since epoch as a string
		numsecs = "%d" % int(time.time())

		md5sum.update(numsecs)					## Seconds and...
		md5sum.update(self.config.rc_md5_key)	## md5 key...
	
		ascii_md5 = ''

		for ch in range(0,len(md5sum.digest())):
			ascii_md5 = ascii_md5 + "%02.2x" % ord(md5sum.digest()[ch])
		
		self.log.write(5, "  generate_auth(): Authentication = %s-%s" % (numsecs, ascii_md5))

		return "%s-%s" % (numsecs, ascii_md5)

	#------------------------------------------------------------------------------
	def __get_auth_tokens(self):
		"""
		Reads and parse the authorization string from the X-ASK-Auth SMTP header.
		Normally, a tuple in the format (int(numsecs),md5sum) is returned. If the header
		does not exist or cannot be parsed, ("","") is returned.
		"""

		## Auth string
		authstring = self.msg.getheader("X-ASK-Auth", "")

		if not authstring:
			self.log.write(1, "  __get_auth_tokens(): No X-ASK-Auth SMTP header found")
			return ("","")

		self.log.write(5, "  __get_auth_tokens(): Authentication string = [%s]" % authstring)

		## Parse age and MD5
		res = re.search("([0-9]*)-([0-9a-f]*)", authstring, re.IGNORECASE)

		if not res:
			self.log.write(1, "  __get_auth_tokens(): Cannot parse X-ASK-Auth SMTP header")
			
		original_time = int(res.group(1))
		original_md5  = res.group(2)

		return(int(original_time), original_md5)

	#------------------------------------------------------------------------------
	def validate_auth_md5(self):
		"""
		Validates the MD5sum part of the X-ASK-Auth SMTP header. 
		Returns 1 if the authentication is valid, zero otherwise.
		"""

		(original_time, original_ascii_md5) = self.__get_auth_tokens()

		if not original_ascii_md5:
			self.log.write(1, "  validate_auth_md5(): Cannot read authorization tokens. Authentication Failed.")

		## Check the md5sum (original_time + rc_md5_key)
		md5sum = md5.new()
		md5sum.update("%s" % original_time)		## Seconds and...
		md5sum.update(self.config.rc_md5_key)	## md5 key...
	
		ascii_md5 = ''

		for ch in range(0,len(md5sum.digest())):
			ascii_md5 = ascii_md5 + "%02.2x" % ord(md5sum.digest()[ch])
		
		## Compare
		if ascii_md5 == original_ascii_md5:
			self.log.write(1, "  validate_auth_md5(): Authentication succeeded")
			return 1
		else:
			self.log.write(1, "  validate_auth_md5(): MD5 tokens do not match (should be %s). Authentication failed" % ascii_md5)
			return 0

	#------------------------------------------------------------------------------
	def validate_auth_time(self, maxdays):
		"""
		Validates the time part of the X-ASK-Auth SMTP header. 
		Returns 1 if the authentication is valid (newer than maxdays), 0 if not.
		"""

		(original_time, original_ascii_md5) = self.__get_auth_tokens()

		if not original_ascii_md5:
			self.log.write(1, "  validate_auth_time(): Cannot read authorization tokens. Authentication Failed.")

		## Check if original_time is older than 'maxdays' days.

		current_time = int(time.time())

		if (original_time >= (current_time - (maxdays * 86400))):
			self.log.write(1, "  validate_auth_time(): Time Authentication succeeded")
			return 1
		else:
			self.log.write(1, "  validate_auth_time(): Auth time is older than %d days. Authentication failed" % maxdays)
			return 0

	#------------------------------------------------------------------------------
	def __contains_string(self, regexp):
		"Returns line containing string if regexp matches any line in the current message"

		self.fh.seek(0)

		while 1:

			buf = self.fh.readline()

			if (buf == ''):
				break
		
			if (re.search(regexp, buf, re.IGNORECASE)):
				return buf
		return ""

	#------------------------------------------------------------------------------
	def __save_to(self, dest):
		"""
		Saves the current message text into file 'dest'
		"""
		
		self.log.write(5, "  __save_to(): Copying to %s" % dest)

		self.fh.seek(0)
		fh_output = open(dest, "w")

		while 1:
			buf = self.fh.readline()

			if buf == '':
				break

			fh_output.write(buf)

		fh_output.close()

	#------------------------------------------------------------------------------
	def __match_recipient(self):
		"""
		This function will do its best to determine which of our rc_mymails was
		the recipient of the current message. If no entry can be found in the
		rc_mymails list that matches at least one of the recipients, the first
		entry (rc_mymails[0]) will be used.
		"""

		for (recipient_name, recipient_mail) in self.get_recipients():
			for our_address in self.config.rc_mymails:
				if recipient_mail == our_address:
					self.log.write(1, "  __match_recipient(): Found a match with %s" % our_address)
					return our_address

		self.log.write(1, "  __match_recipient(): No Match found. Using %s" % self.config.rc_mymails[0])

		return self.config.rc_mymails[0]
		
#------------------------------------------------------------------------------

## EOF ##
